/*
 * InteractiveModel.java
 *
 * Created on 30 mai 2007, 05:09
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package net.library.jiga.sf3.game;

import net.library.jiga.sf3.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.library.jiga.sf3.system.input.Interactive;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.datatransfer.DataFlavor;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintStream;
import java.io.Serializable;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;
import javax.swing.AbstractAction;
import javax.swing.Action;
import net.library.jiga.sf3.system.CoalescedThreadsMonitor;
import net.library.jiga.sf3.system.Console;
import net.library.jiga.sf3.system.ErrorListener;
import net.library.jiga.sf3.system.input.KeyEventWrapper;
import net.java.games.input.*;
import net.library.jiga.sf3.game.physics.Player;

/**
 *
 * @author website
 */
public class InteractiveModel extends Model implements Serializable, Interactive {

    /***/
    private transient Controller gamePad = null;
    /***/
    public static final String _MIME_TYPE = DataFlavor.javaSerializedObjectMimeType;
    /***/
    public static final String[] _MIME_EXT = new String[]{"mc4", "MC4"};
    /***/
    public static final DataFlavor _dataFlavor = new DataFlavor(InteractiveModel.class, "sf3jswing Interactive Model");
    /***/
    private transient Map<Long, KeyEventWrapper> threadedEventsSequence;
    /***/
    private transient SortedMap<Long, KeyEventWrapper> comboKeys;
    /***/
    private transient Map<KeyEventWrapper, Long> pressedKeys;
    /***/
    private transient ConcurrentHashMap<Long, KeyEventWrapper> _threadedEventsSequence;
    /***/
    private transient TreeMap<Long, KeyEventWrapper> _comboKeys;
    /***/
    private transient HashMap<KeyEventWrapper, Long> _pressedKeys;
    /***/
    public transient Map<KeyEventWrapper, String> keyEvents;
    /***/
    private HashMap<KeyEventWrapper, String> _keyEvents;
    /***/
    int highestKeysSequence = 0;
    /***/
    long highestKeysSequenceTiming = KeyEventWrapper.DEFAULT_TIMING;
    /***/
    private static final int HDR_KEYEVENTS = 0;
    /***/
    private static final int HDR_CURRENTACTION = 1;
    /***/
    private transient boolean dispatching = false;
    /***/
    private transient boolean processing = false;
    /***/
    private transient CoalescedThreadsMonitor dispatchMonitor;
    /***/
    private transient CoalescedThreadsMonitor processMonitor;

    /***/
    InteractiveModel(String name, int res, int floor, Dimension spriteDimension, Dimension spriteInsideDimension, int[][] frames, String[] reversedAnims, String[] sfxRsrcs, HashMap<KeyEventWrapper, String> keyEvents) {
        super(name, res, floor, spriteDimension, spriteInsideDimension, frames, reversedAnims, sfxRsrcs);
        this.keyEvents = Collections.synchronizedMap(_keyEvents = keyEvents);
        highestKeysSequence = highestKeysSequence(this.keyEvents);
        highestKeysSequenceTiming = highestKeysSequenceTiming(this.keyEvents);
        enableInputMethods(false);
        resetKeyEventsMaps();
    }

    /***/
    private void resetKeyEventsMaps() {
        pressedKeys = Collections.synchronizedMap(_pressedKeys = new HashMap<KeyEventWrapper, Long>());
        threadedEventsSequence = Collections.synchronizedMap(_threadedEventsSequence = new ConcurrentHashMap<Long, KeyEventWrapper>());
        comboKeys = Collections.synchronizedSortedMap(_comboKeys = new TreeMap<Long, KeyEventWrapper>());
    }

    /***/
    public InteractiveModel(String name, int res, int floor, Dimension spriteDimension, Dimension spriteInsideDimension, int[][] frames, String[] reversedAnims, String[] sfxRsrcs) {
        this(name, res, floor, spriteDimension, spriteInsideDimension, frames, reversedAnims, sfxRsrcs, new HashMap<KeyEventWrapper, String>());
    }

    /***/
    public static int highestKeysSequence(Map<KeyEventWrapper, String> keyEvents) {
        int highestKeysSequence = 0;
        Map<KeyEventWrapper, String> _keyEvents = Collections.synchronizedMap(keyEvents);
        Set<KeyEventWrapper> set = _keyEvents.keySet();
        synchronized (_keyEvents) {
            for (Iterator<KeyEventWrapper> i = set.iterator(); i.hasNext();) {
                int n = i.next()._sequence.size();
                if (n > highestKeysSequence) {
                    highestKeysSequence = n;
                }
            }
        }
        return highestKeysSequence;
    }

    /***/
    public static long highestKeysSequenceTiming(Map<KeyEventWrapper, String> keyEvents) {
        long highestKeysSequenceTiming = KeyEventWrapper.DEFAULT_TIMING;
        Map<KeyEventWrapper, String> _keyEvents = Collections.synchronizedMap(keyEvents);
        Set<KeyEventWrapper> set = _keyEvents.keySet();
        synchronized (_keyEvents) {
            for (Iterator<KeyEventWrapper> i = set.iterator(); i.hasNext();) {
                long n = i.next()._timing;
                if (n > highestKeysSequenceTiming) {
                    highestKeysSequenceTiming = n;
                }
            }
        }
        return highestKeysSequenceTiming;
    }
    /***/
    private transient Thread eventReader = null;
    /***/
    private transient Thread currentAnimSwitch = null;

    /***/
    private Action getKeyEventsAction(int handler) {
        switch (handler) {
            default:
                return null;
            case HDR_CURRENTACTION:
                return new AbstractAction("Action-Model Current Animation") {

                    public void actionPerformed(ActionEvent e) {
                        if (!isLoaded()) {
                            return;
                        }
                        final CoalescedThreadsMonitor monitor1 = animSynch;
                        synchronized (monitor1.getMonitor(false)) {
                            if (currentAnim instanceof Animation) {
                                long playTime = currentAnim.getStartTime();
                                if (System.currentTimeMillis() - playTime > currentAnim.realTimeLength() && playTime > 0) {
                                    if (animName_lock instanceof String) {
                                        anim(animName_lock, frameRate, false, true);
                                    } else {
                                        anim(animsID_default, frameRate, false, false);
                                    }
                                }
                            }
                        }
                        System.out.println("Action-Model's doing " + animName + " action now.");
                    }
                };
            case HDR_KEYEVENTS:
                return new AbstractAction("Action-KeyEvents-Handler_" + getClass().getName()) {

                    public void actionPerformed(ActionEvent e) {
                        if (!isLoaded()) {
                            return;
                        }
                        // match animation troughout keys sequences
                        Map<String, Vector> filtered = filterKeysSequence(System.currentTimeMillis(), true);
                        SortedMap<Long, KeyEventWrapper> unreadKewSequence;
                        Map<Boolean, KeyEventWrapper> match;
                        if ((match = matchKeyEvents(new KeyEventWrapper(filtered.get("event")))).get(true) instanceof KeyEventWrapper) {
                            // switch anim
                            anim(keyEvents.get(match.get(true)), frameRate, true, false);
                            // forward unread _sequence
                            unreadKewSequence = new TreeMap<Long, KeyEventWrapper>();
                            int n = match.get(true)._sequence.size();
                            for (int j = (n > 0) ? n : 1; j < filtered.get("event").size(); j++) {
                                unreadKewSequence.put((Long) filtered.get("time").get(j), (KeyEventWrapper) filtered.get("event").get(j));
                            }
                            sendKeyEventWrapperSequence(unreadKewSequence);
                        } else {
                            // forward unread _sequence
                            unreadKewSequence = new TreeMap<Long, KeyEventWrapper>();
                            for (int j = 0; j < filtered.get("event").size(); j++) {
                                unreadKewSequence.put((Long) filtered.get("time").get(j), (KeyEventWrapper) filtered.get("event").get(j));
                            }
                            sendKeyEventWrapperSequence(unreadKewSequence);
                        }
                        System.out.println("##### REACTIVE MODEL INPUT ***** > unread : " + unreadKewSequence);
                    }
                };
        }
    }
    /***/
    private transient KeyEventWrapper currentCombo;

    /***/
    private Map<String, Vector> filterKeysSequence(long filterTime, boolean grab) {
        System.out.println("#### filter keys");
        Map<String, Vector> filteredKeysSequence = Collections.synchronizedMap(new HashMap<String, Vector>(2));
        Vector<KeyEventWrapper> sequence = new Vector<KeyEventWrapper>();
        Vector<Long> sequenceTime = new Vector<Long>();
        filteredKeysSequence.put("time", sequenceTime);
        filteredKeysSequence.put("event", sequence);
        // read the _sequence
        long now = filterTime;
        long past = now - highestKeysSequenceTiming;
        SortedMap<Long, KeyEventWrapper> sortedEvents = Collections.synchronizedSortedMap(new TreeMap<Long, KeyEventWrapper>());

        // filter keyEvents by time
        Set<Long> set0 = threadedEventsSequence.keySet();
        synchronized (threadedEventsSequence) {
            for (Iterator<Long> it = set0.iterator(); it.hasNext();) {
                Long time = it.next();
                if (time <= now && time > past) {
                    sortedEvents.put(time, threadedEventsSequence.get(time));
                    if (grab) {
                        threadedEventsSequence.remove(time);
                    }
                } else {
                    threadedEventsSequence.remove(time);
                }
            }
        }
        System.out.println("#### keys filtered to read on current time = " + sortedEvents.size());
        // add pressed keys
        synchronized (pressedKeys) {
            for (Iterator<Map.Entry<KeyEventWrapper, Long>> i = pressedKeys.entrySet().iterator(); i.hasNext();) {
                Map.Entry<KeyEventWrapper, Long> entry = i.next();
                sortedEvents.put(entry.getValue(), entry.getKey());
            }
        }
        System.out.println("#### pressed keys added = " + sortedEvents.size());
        // make parallel time and event vectors
        Set<Long> set1 = sortedEvents.keySet();
        synchronized (sortedEvents) {
            for (Iterator<Long> i = set1.iterator(); i.hasNext();) {
                Long key = i.next();
                sequence.add(sortedEvents.get(key));
                sequenceTime.add(key);
                System.out.print(" > " + sequence.lastElement());
            }
        }
        // set combo keys sequence
        for (int i = 0; i < sequenceTime.size(); i++) {
            if (sequence.get(i)._sequence.isEmpty()) {
                comboKeys.put(sequenceTime.get(i), sequence.get(i));
            }
        }
        // add combo keys sequence and reset combo
        synchronized (comboKeys) {
            Vector<KeyEventWrapper> combo = new Vector<KeyEventWrapper>();
            for (Iterator<Long> i = comboKeys.keySet().iterator(); i.hasNext();) {
                Long l = i.next();
                if (l <= now && l > past) {
                    combo.add(comboKeys.get(l));
                } else {
                    i.remove();
                }
            }
            currentCombo = new KeyEventWrapper(combo);
        }
        System.out.println();
        System.out.println("#### filter keys end ");
        return filteredKeysSequence;
    }

    /** @return KeyEventWrapper instance with either sequence of keycodes or keycode */
    private Map<Boolean, KeyEventWrapper> matchKeyEvents(KeyEventWrapper kew) {
        Map<Boolean, KeyEventWrapper> match = null;
        Set<KeyEventWrapper> set2 = keyEvents.keySet();
        synchronized (keyEvents) {
            // find combo (min 2 keyEvents)
            synchronized (comboKeys) {
                System.out.println("current Combo : " + currentCombo);
                for (Iterator<KeyEventWrapper> i = comboKeys.values().iterator(); i.hasNext();) {
                    System.out.print(i.next() + " > ");
                }
                for (Iterator<KeyEventWrapper> i = set2.iterator(); i.hasNext();) {
                    KeyEventWrapper matcher = i.next();
                    if (currentCombo instanceof KeyEventWrapper) {
                        if (!currentCombo._sequence.isEmpty()) {
                            for (int j = 0; j < currentCombo._sequence.size() - 1; j++) {
                                KeyEventWrapper combo = new KeyEventWrapper(new Vector<KeyEventWrapper>(currentCombo._sequence.subList(j, currentCombo._sequence.size())));
                                if (matcher.compareTo(combo) <= 0) {
                                    if (kew != null) {
                                        if (!kew._sequence.isEmpty()) {
                                            for (int k = 0; k < j && k < kew._sequence.size(); k++) {
                                                kew._sequence.remove(k);
                                            }
                                        }
                                    }
                                    System.out.println();
                                    keyEvents.notify();
                                    match = Collections.singletonMap(true, matcher);
                                    break;
                                }
                            }
                            if (match != null) {
                                break;
                            }
                        }
                    }
                }
                System.out.println();
                comboKeys.notifyAll();
            }
            if (match == null) {
                if ((kew == null) ? true : kew.isBlankEvent() || keyEvents.isEmpty()) {
                    match = Collections.singletonMap(false, kew);
                }
            }
            if (match == null) {// compare specified sequence to the Model base sequences
                for (Iterator<KeyEventWrapper> i = set2.iterator(); i.hasNext();) {
                    KeyEventWrapper matcher = i.next();
                    KeyEventWrapper readKew = kew;
                    if (!kew._sequence.isEmpty()) {
                        readKew = new KeyEventWrapper(new Vector(kew._sequence.subList(0, Math.min(kew._sequence.size(), matcher._sequence.size()))));
                    }
                    if (matcher.compareTo(readKew) <= 0) {
                        keyEvents.notify();
                        match = Collections.singletonMap(true, matcher);
                        break;
                    }
                }
            }
            if (match == null) {
                // compare the eldest (first in list) specified keyEvent to the Model base
                for (Iterator<KeyEventWrapper> i = set2.iterator(); i.hasNext();) {
                    KeyEventWrapper matcher = i.next();
                    if (kew._sequence.isEmpty()) {
                        if (matcher.compareTo(kew) <= 0) {
                            keyEvents.notify();
                            match = Collections.singletonMap(true, matcher);
                            break;
                        }
                    } else if (matcher.compareTo(kew._sequence.get(0)) <= 0) {
                        keyEvents.notify();
                        match = Collections.singletonMap(true, matcher);
                        break;
                    }
                }
            }
            keyEvents.notify();
        }
        if (match == null) {
            match = Collections.singletonMap(false, kew);
        } else if (match.get(true) instanceof KeyEventWrapper) {
            System.err.println("ReactiveModel " + id +
                    " has matched " + match.get(true)._sequence.size() +
                    " keys to animate.");
        }
        return match;
    }

    /***/
    private void sendKeyEventWrapperSequence(SortedMap<Long, KeyEventWrapper> unreadKewSequence) {
        SortedMap<Long, KeyEventWrapper> newKew = Collections.synchronizedSortedMap(unreadKewSequence);
        synchronized (newKew) {
            for (Iterator<Long> i = newKew.keySet().iterator(); i.hasNext();) {
                long l = i.next();
                threadedEventsSequence.put(l, newKew.get(l));
            }
        }
        System.out.println("##### event sent");
    }

    /***/
    public void addKeyEventWrapperMapping(KeyEventWrapper kew, String modelKey) {
        keyEvents.put(kew, modelKey);
        highestKeysSequence = highestKeysSequence(keyEvents);
        highestKeysSequenceTiming = highestKeysSequenceTiming(keyEvents);
    }

    /***/
    public void addKeyEventWrapperMap(Map<KeyEventWrapper, String> map) {
        keyEvents.putAll(map);
        highestKeysSequence = highestKeysSequence(keyEvents);
        highestKeysSequenceTiming = highestKeysSequenceTiming(keyEvents);
    }

    /***/
    public void removeKeyEventWrapperMapping(KeyEventWrapper kew) {
        keyEvents.remove(kew);
        highestKeysSequence = highestKeysSequence(keyEvents);
        highestKeysSequenceTiming = highestKeysSequenceTiming(keyEvents);
    }

    /***/
    public void clearKeyEventWrapperMap() {
        keyEvents.clear();
        highestKeysSequence = 0;
        highestKeysSequenceTiming = KeyEventWrapper.DEFAULT_TIMING;
    }

    /** TODO combo keys management*/
    public boolean dispatchKeyEvent(KeyEvent e) {
        int pty = Thread.currentThread().getPriority();
        Thread.currentThread().setPriority(dispatchMonitor.getMaxPriority());
        boolean interrupt_ = false;
        boolean b = false;
        try {
            final CoalescedThreadsMonitor monitor = processMonitor;
            synchronized (monitor.getMonitor(false)) {
                while (processing) {
                    monitor.waitOnMonitor(10);
                }
                monitor.notifyOnMonitor();
                dispatching = true;
                final KeyEvent _e = e;
                SortedMap<Long, KeyEventWrapper> kews = new TreeMap<Long, KeyEventWrapper>();
                Set<KeyEventWrapper> set = pressedKeys.keySet();
                long now = e.getWhen();
                // manage key pression time
                switch (e.getID()) {
                    case KeyEvent.KEY_PRESSED:
                        synchronized (pressedKeys) {
                            for (Iterator<KeyEventWrapper> i = set.iterator(); i.hasNext();) {
                                KeyEventWrapper kew = i.next();
                                if (kew instanceof KeyEventWrapper) {
                                    if (kew._keyCode == e.getKeyCode()) {
                                        b = true;
                                        break;
                                    }
                                }
                            }
                            if (b) {
                                break;
                            }
                        }
                        System.out.println("#### dispatch new key pressed");
                        pressedKeys.put(new KeyEventWrapper(e.getKeyCode(), KeyEvent.KEY_PRESSED), now);
                        break;
                    case KeyEvent.KEY_RELEASED:
                        synchronized (pressedKeys) {
                            for (Iterator<KeyEventWrapper> i = set.iterator(); i.hasNext();) {
                                KeyEventWrapper kew = i.next();
                                if (kew instanceof KeyEventWrapper) {
                                    if (kew._keyCode == e.getKeyCode()) {
                                        kews.put(now, new KeyEventWrapper(e.getKeyCode(), KeyEvent.KEY_PRESSED, now - pressedKeys.get(kew)));
                                        i.remove();
                                        break;
                                    }
                                }
                            }
                        }
                        kews.put(now + 1, new KeyEventWrapper(e.getKeyCode(), KeyEvent.KEY_RELEASED));
                        break;
                    default:
                        b = false;
                        break;
                }
                sendKeyEventWrapperSequence(kews);
                // match animation troughout keys sequences
                Map<String, Vector> filtered = filterKeysSequence(now, false);
                SortedMap<Long, KeyEventWrapper> unreadKewSequence;
                Map<Boolean, KeyEventWrapper> match;
                b = (match = matchKeyEvents(new KeyEventWrapper(filtered.get("event")))).get(true) instanceof KeyEventWrapper;
            }
        } catch (InterruptedException ex) {
            ex.printStackTrace();
            interrupt_ = true;
        } finally {
            final CoalescedThreadsMonitor monitor = dispatchMonitor;
            synchronized (monitor.getMonitor(false)) {
                dispatching = false;
                monitor.notifyAllOnMonitor();
            }
            if (interrupt_) {
                return false;
            }
            Thread.currentThread().setPriority(pty);
            return b;
        }
    }

    /***/
    private void writeObject(ObjectOutputStream out) throws IOException {
        out.defaultWriteObject();
    }

    /***/
    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        keyEvents = Collections.synchronizedMap(_keyEvents);
        gamePad = null;
        resetKeyEventsMaps();
        System.out.println(toString());
    }

    /***/
    public String toString() {
        String s = super.toString();
        s += " ReactiveModel for model " + id + " keys mapping : { ";
        synchronized (keyEvents) {
            for (Iterator<KeyEventWrapper> i = keyEvents.keySet().iterator(); i.hasNext();) {
                KeyEventWrapper kew = i.next();
                s += " [" + kew.toString() + " <-> " + keyEvents.get(kew) + "] ";
            }
        }
        s += " }" + Console.newLine;
        return s;
    }

    /***/
    public void setGamePadController(Controller gamePad) {
        this.gamePad = gamePad;
    }

    /***/
    public Controller getGamePadController() {
        return gamePad;
    }

    /***/
    private SortedMap<Long, KeyEventWrapper> processGamepadEvent(String key, int ID, long time, SortedMap<Long, KeyEventWrapper> seq) {
        if (ID == KeyEvent.KEY_PRESSED) {
            synchronized (pressedKeys) {
                for (Iterator<KeyEventWrapper> it = pressedKeys.keySet().iterator(); it.hasNext();) {
                    KeyEventWrapper kew = it.next();
                    if (kew instanceof KeyEventWrapper) {
                        if (kew._keyCode == Player._keyCodeMap.get(key)) {
                            return seq;
                        }
                    }
                }
            }
            pressedKeys.put(new KeyEventWrapper(Player._keyCodeMap.get(key), KeyEvent.KEY_PRESSED), time);
        } else if (ID == KeyEvent.KEY_RELEASED) {
            synchronized (pressedKeys) {
                for (Iterator<KeyEventWrapper> it = pressedKeys.keySet().iterator(); it.hasNext();) {
                    KeyEventWrapper kew = it.next();
                    if (kew instanceof KeyEventWrapper) {
                        if (kew._keyCode == Player._keyCodeMap.get(key)) {
                            seq.put(time, new KeyEventWrapper(Player._keyCodeMap.get(key), KeyEvent.KEY_PRESSED, time - pressedKeys.get(kew)));
                            it.remove();
                            break;
                        }
                    }
                }
            }
            seq.put(time + 1, new KeyEventWrapper(Player._keyCodeMap.get(key), KeyEvent.KEY_RELEASED));
        }
        return seq;

    }

    /***/
    public void processKeyEventsStack() {
        try {
            Thread t = currentAnimSwitch = new Thread(processMonitor, new Runnable() {

                public void run() {
                    boolean interrupt_ = false;
                    try {
                        final CoalescedThreadsMonitor monitor0 = dispatchMonitor;
                        synchronized (monitor0.getMonitor(true)) {
                            while (dispatching) {
                                monitor0.waitOnMonitor(10);
                            }
                            monitor0.notifyOnMonitor();
                            processing = true;
                            if (gamePad instanceof Controller) {
                                Component xAxis = gamePad.getComponent(Component.Identifier.Axis.X);
                                Component yAxis = gamePad.getComponent(Component.Identifier.Axis.Y);
                                float lastPollX = Math.round(xAxis.getPollData());
                                float lastPollY = Math.round(yAxis.getPollData());
                                gamePad.poll();
                                EventQueue eq = gamePad.getEventQueue();
                                Event e = new Event();
                                SortedMap<Long, KeyEventWrapper> seq = new TreeMap<Long, KeyEventWrapper>();
                                while (eq.getNextEvent(e)) {
                                    long time = (long) ((float) e.getNanos() / 1000f);
                                    float val = Math.round(e.getValue());
                                    Component button = e.getComponent();
                                    int ID = (val == -1f || val == 1f) ? KeyEvent.KEY_PRESSED : KeyEvent.KEY_RELEASED;
                                    if (button instanceof Component.Identifier.Axis) {
                                        String direction = "";
                                        if (Component.Identifier.Axis.X.getName().equals(button.getName())) {
                                            if (new String("ONE").equals(attributes.get("player"))) {
                                                direction = (val == -1f) ? "backwards" : ((val == 1f) ? "forwards" : ((lastPollX == -1f) ? "backwards" : ((lastPollX == 1f) ? "forwards" : "")));
                                            } else if (new String("TWO").equals(attributes.get("player"))) {
                                                direction = (val == 1f) ? "backwards" : ((val == -1f) ? "forwards" : ((lastPollX == 1f) ? "backwards" : ((lastPollX == -1f) ? "forwards" : "")));
                                            }
                                            lastPollX = val;
                                        } else if (Component.Identifier.Axis.Y.getName().equals(button.getName())) {
                                            direction = (val == 1f) ? "down" : ((val == -1f) ? "up" : ((lastPollY == 1f) ? "down" : ((lastPollY == -1f) ? "up" : "")));
                                            lastPollY = val;
                                        }
                                        if (direction.length() > 0) {
                                            seq = processGamepadEvent(attributes.get("player") + "-" + direction, ID, time, seq);
                                        }
                                    } else {
                                        synchronized (Player._gamepadCodeMap) {
                                            for (Iterator<String> i = Player._gamepadCodeMap.keySet().iterator(); i.hasNext();) {
                                                String key = i.next();
                                                if (key.startsWith((String) attributes.get("player"))) {
                                                    String name = Player._gamepadCodeMap.get(key);
                                                    if (button.getName().equals(name)) {
                                                        seq = processGamepadEvent(key, ID, time, seq);
                                                        break;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                sendKeyEventWrapperSequence(seq);
                            }
                            ActionEvent ae = new ActionEvent(InteractiveModel.this, (int) System.currentTimeMillis(), "currentAction");
                            getKeyEventsAction(HDR_CURRENTACTION).actionPerformed(ae);
                            getKeyEventsAction(HDR_KEYEVENTS).actionPerformed(ae);
                        }
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                        interrupt_ = true;
                    } finally {
                        final CoalescedThreadsMonitor monitor0 = processMonitor;
                        synchronized (monitor0.getMonitor(false)) {
                            processing = false;
                            monitor0.notifyAllOnMonitor();
                        }
                        if (interrupt_) {
                            Thread.currentThread().interrupt();
                        }
                    }
                }
            }, "T-currentAnimSwitch");
            t.setPriority(Thread.MAX_PRIORITY);
            t.start();
            t.join();
        } catch (InterruptedException ex) {
            ex.printStackTrace();
            Thread.currentThread().interrupt();
        }
    }

    /***/
    protected void paintComponent(Graphics g) {
        processKeyEventsStack();
        super.paintComponent(g);
    }

    /***/
    @Override
    public Object clearResource() {
        if (dispatchMonitor instanceof ThreadGroup) {
            dispatchMonitor.interrupt();
        }
        if (processMonitor instanceof ThreadGroup) {
            processMonitor.interrupt();
        }
        return super.clearResource();
    }

    /***/
    public void setThreadGroup(ThreadGroup tg) {
        super.setThreadGroup(tg);
        processing = false;
        dispatching = false;
        if (dispatchMonitor instanceof ThreadGroup) {
            dispatchMonitor.interrupt();
        }
        if (processMonitor instanceof ThreadGroup) {
            processMonitor.interrupt();
        }
        if (tg instanceof ThreadGroup) {
            dispatchMonitor = new CoalescedThreadsMonitor(tg, "TG-" + getClass().getName() + "-dispatch " + id);
            processMonitor = new CoalescedThreadsMonitor(tg, "TG-" + getClass().getName() + "-process " + id);
        } else {
            dispatchMonitor = new CoalescedThreadsMonitor("TG-" + getClass().getName() + "-dispatch " + id);
            processMonitor = new CoalescedThreadsMonitor("TG-" + getClass().getName() + "-process " + id);
        }
        dispatchMonitor.setMaxPriority(Thread.MAX_PRIORITY);
        dispatchMonitor.setCoalesceEnabled(false);
        processMonitor.setMaxPriority(Thread.MAX_PRIORITY);
        processMonitor.setCoalesceEnabled(false);
    }
}
