/*
 * Model.java
 *
 * Created on 30 mai 2007, 05:08
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package net.library.jiga.sf3.game;

import net.library.jiga.sf3.*;
import java.awt.datatransfer.DataFlavor;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.library.jiga.installer.UIMessage;
import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Component;
import java.awt.Composite;
import java.awt.Dimension;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.MediaTracker;
import java.awt.Paint;
import java.awt.Point;
import java.awt.geom.AffineTransform;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.RandomAccessFile;
import java.io.Serializable;
import java.net.URISyntaxException;
import java.util.Collections;
import java.util.HashMap;
import java.util.SortedMap;
import javax.activation.FileTypeMap;
import javax.activation.MimetypesFileTypeMap;
import javax.media.jai.GraphicsJAI;
import javax.media.jai.PerspectiveTransform;
import javax.swing.JFrame;
import javax.swing.JProgressBar;
import net.library.jiga.installer.ExampleFileFilter;
import net.library.jiga.sf3.system.CoalescedThreadsMonitor;
import net.library.jiga.sf3.system.Resource;
import net.library.jiga.sf3.system.SfMediaTracker;
import net.library.jiga.sf3.system.SpritesCacheAdapter;
import net.library.jiga.sf3.system.SpritesCacheListener;
import net.library.jiga.sf3.system.SpritesCacheManager;
import net.library.jiga.sf3.system.Threaded;

/**
 *
 * @author website
 * A model performs all operations to display itself on Graphics. It manages its own memory usage.*/
public class Model extends Custom implements Serializable, CompositeCapable, Resource, Threaded {

    /***/
    HashMap<String, Serializable> attributes = new HashMap<String, Serializable>();
    /***/
    int res;
    /***/
    public boolean _innerRsrcMode = true;
    /***/
    SpritesCacheManager<Integer, Animation> spm;
    /***/
    private transient MediaTracker mt;
    /***/
    private long id0 = 0L;
    /***/
    transient SortedMap<Integer, Animation> cache;
    /***/
    private Dimension displaySize;
    /***/
    public boolean _paintStatusEnabled = false;
    /***/
    private int mirror = Animation.NONE;
    /***/
    private boolean transform = true;
    /***/
    private AffineTransform tx = new AffineTransform();
    /***/
    private PerspectiveTransform perspective = new PerspectiveTransform();
    /***/
    protected long frameRate = (long) (1000f / 24f);
    /***/
    protected Animation currentAnim;
    /***/
    protected boolean reversed = false;
    /***/
    protected boolean compositeEnabled = false;
    /***/
    protected transient Paint pnt = null;
    /***/
    protected transient Color clr = null;
    /***/
    protected transient Composite cps = null;
    /***/
    public int _type = Sprite.DEFAULT_TYPE;
    /***/
    protected transient boolean validating = false;
    /***/
    protected transient boolean painting = false;
    /***/
    protected transient CoalescedThreadsMonitor animSynch;
    /***/
    protected transient CoalescedThreadsMonitor validateMonitor;
    /***/
    protected transient CoalescedThreadsMonitor paintMonitor;
    /***/
    public final static int LOADED = 0;
    /***/
    public final static int LOADING = 1;
    /***/
    public final static int UNLOADED = 2;
    protected transient String animName_lock;
    /***/
    int state = UNLOADED;

    public MediaTracker getMt() {
        return mt;
    }

    public void setMt(MediaTracker mt, Component obs) {
        final CoalescedThreadsMonitor monitor = animSynch;
        synchronized (monitor.getMonitor(false)) {
            this.mt = mt;
            this.obs = obs;
            monitor.notifyOnMonitor();
        }
    }

    /***/
    public Model(String path,
            int res,
            int floor,
            Dimension spriteDimension,
            Dimension spriteInsideDimension,
            int[][] frames,
            String[] reversedAnims,
            String[] sfxRsrcs) {
        super(path, res, floor, spriteDimension, spriteInsideDimension, frames, reversedAnims, sfxRsrcs);
        mt = new SfMediaTracker(obs = this);
        displaySize = new Dimension(getWidth(res), getHeight(res));
        spm = new SpritesCacheManager<Integer, Animation>(frames.length);
        cache = Collections.synchronizedSortedMap(spm);
        setEnabled(false);
        spm.setSwapDiskEnabled(true);
        spm.setAutoCleanupEnabled(false);
        spm.setCompressionEnabled(false);
        setThreadGroup(new CoalescedThreadsMonitor("Model-TG"));
    }

    /***/
    protected void finalize() throws Throwable {
        paintMonitor.cancelAll(true);
        validateMonitor.cancelAll(true);
        animSynch.cancelAll(true);
        UIMessage.killProgress(id0);
        super.finalize();
    }

    /***/
    public Dimension getInsetSize() {
        return spriteSize;
    }

    /***/
    public void setResolution(int res) {
        this.res = res;
    }

    /***/
    public int getResolution() {
        return res;
    }

    /***/
    public int getFloorDiff() {
        return (int) ((float) getFloorDiff(getResolution()) * (float) getDisplaySize().height / (float) getHeight(getResolution()));
    }

    /***/
    public Serializable setAttribute(String name, Serializable value) {
        return attributes.put(name, value);
    }

    /***/
    public Serializable getAttribute(String name) {
        return attributes.get(name);
    }

    /***/
    public void setFrameRate(long frameRate) {
        this.frameRate = frameRate;
    }

    /***/
    public long getFrameRate() {
        return frameRate;
    }

    /***/
    public void setCompressionEnabled(boolean b) {
        spm.setCompressionEnabled(b);
    }

    /***/
    public boolean isCompressionEnabled() {
        return spm.isCompressedCacheEnabled();
    }

    /***/
    public void setSwapDiskEnabled(boolean b) {
        spm.setSwapDiskEnabled(b);
    }

    /***/
    public boolean isSwapDiskEnabled() {
        return spm.isSwapDiskEnabled();
    }

    /***/
    public boolean isFlipEnabled() {
        return mirror != Animation.NONE;
    }

    /***/
    public void setFlipEnabled(boolean b, int orientation) {
        mirror = (b) ? orientation : Animation.NONE;
    }

    /***/
    public boolean isZoomEnabled() {
        return zoom != 1.0;
    }
    /***/
    private double zoom = 1.0;

    /***/
    public void setZoomEnabled(boolean b, double zoom) {
        this.zoom = (b) ? zoom : 1.0;
    }

    /***/
    public boolean isTransformEnabled() {
        return transform;
    }

    /***/
    public void setTransformEnabled(boolean b) {
        transform = b;
    }

    /***/
    public void setTx(AffineTransform tx) {
        this.tx = tx;
    }

    /***/
    public AffineTransform getTx() {
        return tx;
    }

    /***/
    public void setPtx(PerspectiveTransform ptx) {
        perspective = ptx;
    }

    /***/
    public PerspectiveTransform getPtx() {
        return perspective;
    }
    /***/
    private transient Component obs;

    /***/
    protected void paintComponent(Graphics g) {
        try {
            draw(this, (Graphics2D) g, 0, 0, false, true, null);
        } catch (InterruptedException e) {
            e.printStackTrace();
            return;
        }
    }
    /***/
    private boolean opaque = false;

    /***/
    public boolean isOpaque() {
        return opaque;
    }

    /***/
    public void setOpaque(boolean b) {
        opaque = b;
    }

    /***/
    public void update(Graphics g) {
        validate();
        if (isOpaque()) {
            g.clearRect(0, 0, getWidth(), getHeight());
        }
        paint(g);
    }

    /***/
    public boolean draw(Component pobs, Graphics2D g, int x, int y, boolean shadowMirrorEnabled, boolean shadowEnabled, Point shadowPos) throws InterruptedException {
        return draw(pobs, g, x, y, shadowMirrorEnabled, shadowEnabled, shadowPos, null, null);
    }

    /***/
    public boolean draw(Component pobs, Graphics2D g, int x, int y, boolean shadowMirrorEnabled, boolean shadowEnabled, Point shadowPos, AffineTransform tx, PerspectiveTransform ptx) throws InterruptedException {
        if (!isLoaded()) {
            return false;
        }
        if (shadowPos == null) {
            shadowPos = new Point(0, 0);
        }
        obs = pobs;
        boolean complete = true;
        int pty = Thread.currentThread().getPriority();
        Thread.currentThread().setPriority(paintMonitor.getMaxPriority());
        spm.ensureListCapacity(spm.getInitialListCapacity());
        GraphicsJAI g2 = Sprite.createGraphicsJAI(g, obs);
        boolean interrupt_ = false;
        try {
            final CoalescedThreadsMonitor monitor0 = validateMonitor;
            synchronized (monitor0.getMonitor(false)) {
                if (validating) {
                    System.err.println(id + " unpacked Model's waitin'...");
                }
                while (validating) {
                    System.err.print(".");
                    monitor0.waitOnMonitor(10);
                }
                painting = true;
                System.err.print("OK\n\r");
                final CoalescedThreadsMonitor monitor1 = animSynch;
                synchronized (monitor1.getMonitor(false)) {
                    g2.translate(x, y);
                    Animation anim = currentAnim;
                    if (anim instanceof Animation) {
                        Sprite shadow = null;
                        AffineTransform shadow_tx = new AffineTransform();
                        if (shadowEnabled) {
                            Sprite sp = (Sprite) spm.memorySensitiveCallback("getCurrentSprite", anim, new Object[]{}, new Class[]{});
                            shadow = (Sprite) spm.memorySensitiveCallback("clone", sp, new Object[]{}, new Class[]{});
                            shadow.setMt(mt, obs);
                        }
                        complete = (Boolean) spm.memorySensitiveCallback("draw", anim, new Object[]{obs, g2, tx, ptx}, new Class[]{Component.class, Graphics2D.class, AffineTransform.class, PerspectiveTransform.class});
                        if (shadowEnabled && shadow instanceof Sprite) {
                            spm.buffer(shadow);
                            shadow.setCompositeEnabled(true);
                            shadow.setPreferredSize(displaySize);
                            shadow.setSize(displaySize);
                            if (shadowMirrorEnabled) {
                                Paint shadow_gradient = new GradientPaint(0, 0, new Color(1.0f, 1.0f, 1.0f, 1.0f), 0, (float) displaySize.height * 0.66f, new Color(1.0f, 1.0f, 1.0f, 0.0f), false);
                                spm.buffer(shadow_gradient);
                                shadow.setPaint(shadow_gradient);
                                shadow.setComposite(AlphaComposite.getInstance(AlphaComposite.DST_IN, 1.0f));
                            } else {
                                shadow_tx.scale(1.0, 0.5);
                                shadow.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_IN, 1.0f));
                                shadow.setColor(Color.BLACK);
                            }
                            shadow.setTX(shadow_tx);
                            shadow.setFlipEnabled(true, Sprite.VERTICAL);
                            shadow.validate();
                            Composite cps = g2.getComposite();
                            float[][] scale = new float[][]{
                                {(float) anim.getWidth() / (float) getWidth(res), (float) anim.getHeight() / (float) getHeight(res)},
                                {0.0f, 0.0f}};
                            scale[1][0] = (float) shadow.getWidth() / (float) getWidth(res);
                            scale[1][1] = (float) shadow.getHeight() / (float) getHeight(res);
                            System.out.println("shadow current scale from original " + shadow.getSize() + " : " + scale[1][0] + "," + scale[1][1]);
                            System.out.println(String.valueOf((int) (-anim.getHeight() + (float) getFloorDiff(res) * (scale[1][1] + scale[0][1]))));
                            g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.5f));
                            g2.translate((double) shadowPos.x + 0.0, (double) shadowPos.y + (double) ((float) anim.getHeight() - (float) getFloorDiff(res) * (scale[1][1] + scale[0][1])));
                            complete = complete && (Boolean) spm.memorySensitiveCallback("draw", shadow, new Object[]{obs, g2, tx, ptx}, new Class[]{Component.class, Graphics2D.class, AffineTransform.class, PerspectiveTransform.class});
                            g2.setComposite(cps);
                            g2.translate(-((double) shadowPos.x + 0.0), -((double) shadowPos.y + (double) ((float) anim.getHeight() - (float) getFloorDiff(res) * (scale[1][1] + scale[0][1]))));
                        }
                        if (_paintStatusEnabled) {
                            String str = "Animation " + animName + " FPS = " + String.valueOf(anim.getFPS());
                            g2.drawString(str, 10, 10);
                        }
                        monitor1.notifyOnMonitor();
                    }
                    g2.translate(-x, -y);
                }
                monitor0.notifyOnMonitor();
            }
        } catch (InterruptedException e) {
            e.printStackTrace();
            interrupt_ = true;
        } catch (Throwable t) {
            t.printStackTrace();
            interrupt_ = true;
        } finally {
            final CoalescedThreadsMonitor monitor1 = paintMonitor;
            synchronized (monitor1.getMonitor(false)) {
                painting = false;
                monitor1.notifyAllOnMonitor();
            }
            spm.ensureListCapacity(spm.getInitialListCapacity());
            Thread.currentThread().setPriority(pty);
            if (interrupt_) {
                throw new InterruptedException("Model " + id + " caught an interruption.");
            }
            return complete;
        }
    }

    /***/
    public String getFPS() {
        final CoalescedThreadsMonitor monitor1 = animSynch;
        synchronized (monitor1.getMonitor(false)) {
            monitor1.notifyOnMonitor();
            if (currentAnim instanceof Animation) {
                return currentAnim.getFPS();
            } else {
                return "0.0";
            }
        }
    }

    /** @param wait if true, this method blocks until data is loaded */
    public void load(boolean wait, boolean showProgress, SpritesCacheListener spml) throws InterruptedException {
        final boolean _showProgress = showProgress;
        final SpritesCacheListener _spml = spml;
        Thread t = new Thread(validateMonitor, new Runnable() {

            public void run() {
                try {
                    load(_showProgress, _spml);
                } catch (InterruptedException ex) {
                    ex.printStackTrace();
                    Thread.currentThread().interrupt();
                    return;
                }
            }
        }, "T-Model-load");
        t.start();
        if (wait) {
            t.join();
        }
    }

    /***/
    private void load(boolean pshowProgress, SpritesCacheListener pspml) throws InterruptedException {
        final boolean showProgress = pshowProgress;
        final SpritesCacheListener spml = pspml;
        Thread t = new Thread(validateMonitor, new Runnable() {

            public void run() {
                boolean interrupt_ = false;
                System.err.print("OK\n\r");
                if (state == LOADING) {
                    System.err.println("Model is already loading. Please wait !");
                    Thread.currentThread().interrupt();
                }
                state = LOADING;
                spm.addSpritesCacheListener(spml);
                if (showProgress) {
                    id0 = UIMessage.displayProgress(0, frames.length, null);
                } else {
                    id0 = UIMessage.newProgress(0, frames.length, null);
                }
                UIMessage.showProgress(id0);
                for (int i = 0; i < frames.length; i++) {
                    Animation animation = null;
                    if (cache.containsKey(i)) {
                        animation = cache.get(i);
                    } else {
                        try {
                            int[] anim = frames[i];
                            animation = new Animation(id + ((_innerRsrcMode) ? "/" : File.separator) + RES_PATHS_map.get(res), _innerRsrcMode, anim[0], anim[1], prefix, suffix + RES_EXT_map.get(res), RES_MIMETYPES_map.get(res), new Dimension(getWidth(res), getHeight(res)));
                            refreshAnimProperties(animation, displaySize, mt, reversed, (int) frameRate, tx, mirror, transform, cps, clr, pnt, _type, zoom);
                            if (showProgress) {
                                animation.addIIOReadProgressListener(new IIOReadProgressAdapter(true));
                            }
                            animation.setCompressedCacheEnabled(spm.isCompressedCacheEnabled());
                            cache.put(i, animation);
                        } catch (URISyntaxException ex) {
                            ex.printStackTrace();
                        }
                    }
                    if (animation != null) {
                        animation.clearResource();
                    }
                    spm.ensureListCapacity(spm.getInitialListCapacity());
                    UIMessage.updateProgress(id0, i, frames.length);
                }
                UIMessage.killProgress(id0);
                state = LOADED;
            }
        });
        t.setPriority(Thread.MAX_PRIORITY);
        t.start();
        t.join();
    }

    /***/
    public double allocSize() {
        return spm.allocSize();
    }

    /***/
    private void refreshAnimProperties(
            Animation anim,
            Dimension displaySize,
            MediaTracker mt,
            boolean reversed,
            int frameRate,
            AffineTransform tx,
            int mirror,
            boolean transformEnabled,
            Composite cps,
            Color clr,
            Paint pnt,
            int type,
            double zoom) {
        /*if(anim.getThreadGroup() != getThreadGroup())
        anim.setThreadGroup(getThreadGroup());*/
        anim.setPreferredSize(displaySize);
        anim.setSize(displaySize);
        anim.setTX(tx);
        anim.setPTX(perspective);
        anim.setMt(mt, obs);
        anim.setTransformEnabled(transformEnabled);
        anim.setFlipEnabled(true, mirror);
        anim.setZoomEnabled(true, zoom);
        anim.setReverseEnabled(reversed);
        anim.setFrameRate((int) frameRate);
        anim.setCompositeEnabled(compositeEnabled);
        anim.setComposite(cps);
        anim.setColor(clr);
        anim.setPaint(pnt);
        anim._type = _type;
        anim.setOpaque(opaque);
    }
    /***/
    protected transient String animName;

    /***/
    public JProgressBar getProgress() {
        JProgressBar jpb = (id0 != 0L) ? UIMessage.getProgressBar(id0) : null;
        if (jpb != null) {
            jpb.setSize(new Dimension(getDisplaySize().width, 30));
        }
        return jpb;
    }

    /***/
    public String getAnimName() {
        return animName;
    }

    /***/
    public boolean equals(Object o) {
        if (super.equals(o)) {
            if (o instanceof Model) {
                return frames.equals(((Model) o).frames);
            }
        }
        return false;
    }

    /***/
    public void validate() {
        super.validate();
        if (!isLoaded()) {
            return;
        }
        final CoalescedThreadsMonitor monitor1 = animSynch;
        synchronized (monitor1.getMonitor(false)) {
            if (currentAnim == null && isLoaded()) {
                anim(animsID_default, frameRate, false, false);
            }
            Animation anim = currentAnim;
            if (anim instanceof Animation) {
                refreshAnimProperties(
                        anim,
                        displaySize,
                        mt,
                        reversed,
                        (int) frameRate,
                        tx,
                        mirror,
                        transform,
                        cps,
                        clr,
                        pnt,
                        _type, zoom);
            }
            monitor1.notifyOnMonitor();
        }
    }
    /***/
    private transient Thread anim = null;

    /***/
    public String getAnimName_Lock() {
        return animName_lock;
    }
    /***/
    public void anim(String animName, long pframeRate, boolean pforce, boolean lock) {
        anim(anim(animName), animName, pframeRate, isReversedAnim(animName), pforce, lock);
    }

    /***/
    private void anim(int pkey, String pname, long pframeRate, boolean preversed, boolean pforce, boolean plock) {
        if (state != LOADED) {
            return;
        }
        if (anim instanceof Thread && !pforce) {
            if (anim.isAlive()) {
                return;
            }
        }
        final int key = pkey;
        final long frameRate = pframeRate;
        final boolean reversed = preversed;
        final boolean lock = plock;
        final String name = (pname == null)?animsID_default:pname;
        final boolean force = pforce;
        Thread t = anim = new Thread(validateMonitor, new Runnable() {

            public void run() {
                if (paintMonitor == null) {
                    return;
                }
                boolean interrupt_ = false;
                final CoalescedThreadsMonitor monitor0 = paintMonitor;
                try {
                    synchronized (monitor0.getMonitor(!force)) {
                        if (painting) {
                            System.err.println("ReactiveModel's " + id + " waitin' for painting...");
                        }
                        while (painting) {
                            System.err.print(".");
                            monitor0.waitOnMonitor(10);
                        }
                        validating = true;
                        boolean breakLoop = false;
                        if (!lock) {
                            if (animName_lock instanceof String) {
                                if (!animName_lock.equals(name) && !name.startsWith(animName_lock)) {
                                    System.out.println(Model.class.getName() + " is locked on " + animName_lock);
                                    breakLoop = true;
                                }
                            }
                        }
                        if (!breakLoop) {
                            //try{
                            System.err.print("OK\n\r");
                            Animation anim = null;
                            if (cache.containsKey(key)) {
                                anim = cache.get(key);
                            }
                            if (anim == null) {
                                anim = new Animation(3, RES_MIMETYPES_map.get(res), displaySize);
                            }
                            refreshAnimProperties(
                                    anim,
                                    displaySize,
                                    mt,
                                    reversed,
                                    (int) frameRate,
                                    tx,
                                    mirror,
                                    transform,
                                    cps,
                                    clr,
                                    pnt,
                                    _type, zoom);
                            final CoalescedThreadsMonitor monitor1 = animSynch;
                            synchronized (monitor1.getMonitor(false)) {
                                if (currentAnim != null) {
                                    if (!currentAnim.equals(anim)) {
                                        currentAnim.stop();
                                    }
                                }
                                anim.play();
                                currentAnim = anim;
                                animName = name;
                                if (lock) {
                                    animName_lock = name;
                                } else if(name.equals(animName_lock)){
                                    animName_lock = null;
                                }
                                Model.this.reversed = reversed;
                                Model.this.frameRate = frameRate;
                                monitor1.notifyOnMonitor();
                            }
                        }
                        monitor0.notifyOnMonitor();
                    }
                } catch (InterruptedException e) {
                    e.printStackTrace();
                    interrupt_ = true;
                } finally {
                    final CoalescedThreadsMonitor monitor1 = validateMonitor;
                    synchronized (monitor1.getMonitor(false)) {
                        validating = false;
                        monitor1.notifyAllOnMonitor();
                    }
                    if (interrupt_) {
                        Thread.currentThread().interrupt();
                        return;
                    }
                }
            }
        }, "T-Model anim-pack");
        t.setPriority(Thread.MAX_PRIORITY);
        t.setDaemon(false);
        t.start();
    /*if(pforce) {
    t.join();
    }*/
    }

    /***/
    public void setDisplaySize(Dimension size) {
        displaySize = size;
    }

    /***/
    public Dimension getDisplaySize() {
        return displaySize;
    }

    /***/
    public Object clearResource() {
        state = UNLOADED;
        if (validateMonitor instanceof ThreadGroup) {
            validateMonitor.interrupt();
        }
        if (paintMonitor instanceof ThreadGroup) {
            paintMonitor.interrupt();
        }
        spm.clear();
        return null;
    }

    /***/
    private void writeObject(ObjectOutputStream out) throws IOException {
        out.defaultWriteObject();
    }

    /***/
    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        cache = Collections.synchronizedSortedMap(spm);
        mt = new SfMediaTracker(obs = this);
        setThreadGroup(new CoalescedThreadsMonitor("Model-TG"));
    }

    /***/
    public boolean isLoading() {
        return (state == LOADING) ? true : false;
    }

    /***/
    public boolean isLoaded() {
        return (state == LOADED) ? true : false;
    }

    /***/
    public int status() {
        return state;
    }
    /***/
    public static final String _MIME_TYPE = DataFlavor.javaSerializedObjectMimeType;
    /***/
    public static final String[] _MIME_EXT = new String[]{"mc3", "MC3"};
    /***/
    public static final DataFlavor _dataFlavor = new DataFlavor(Model.class, "sf3jswing Model");

    public static void _storeMIMETYPES() {
        String s = _MIME_TYPE;
        for (String ext : _MIME_EXT) {
            s += " " + ext;
        }
        if (!((MimetypesFileTypeMap) FileTypeMap.getDefaultFileTypeMap()).getContentType("file." + _MIME_EXT[0]).equals(_MIME_TYPE)) {
            ((MimetypesFileTypeMap) FileTypeMap.getDefaultFileTypeMap()).addMimeTypes(s);
        }
    }

    static {
        _storeMIMETYPES();
    }

    /***/
    public void exportModel(String filename, String[] exts) {
        // once loaded, write model
        ObjectOutputStream oos = null;
        FileOutputStream fos = null;
        ObjectInputStream ois = null;
        FileInputStream fis = null;
        Model mc3 = null;
        if (filename instanceof String) {
            boolean valid = false;
            for (String ext : exts) {
                valid = valid || filename.endsWith(ext);
            }
            if (!valid) {
                filename += "." + exts[0];
            }
        }
        JFrame wait = UIMessage.displayWaiting("writing model to " + filename, null);
        try {
            File old = new File(filename);
            RandomAccessFile raf = new RandomAccessFile(filename, "rws");
            raf.getChannel().truncate(0);
            fos = new FileOutputStream(raf.getFD());
            //FileLock fl = fos.getChannel().lock();
            oos = new ObjectOutputStream(fos);
            // write sprites cache
            oos.writeObject(this);
            //fl.release();
            oos.close();
            fos.close();
        } catch (Exception ex2) {
            ex2.printStackTrace();
            new UIMessage(true, "Model may not be stored to " + filename + " ! error: " + ex2.getMessage(), null);
        } finally {
            // sprite should have stored so we try to recover it from path
            mc3 = importModel(filename);
            wait.dispose();
            if (mc3 instanceof Model) {
                new UIMessage(true, "Model's been successfully stored to " + filename + " !", null);
            }
        }
    }

    /***/
    public static Model importModel(String filename) {
        ObjectInputStream ois = null;
        FileInputStream fis = null;
        Model mc3 = null;
        JFrame wait = UIMessage.displayWaiting("loading model " + filename, null);
        try {
            fis = new FileInputStream(new RandomAccessFile(filename, "r").getFD());
            //FileLock fl = fis.getChannel().lock();
            ois = new ObjectInputStream(fis);
            mc3 = (Model) ois.readObject();
        //fl.release();
        } catch (Exception ex) {
            ex.printStackTrace();
        } finally {
            if (ois != null) {
                try {
                    ois.close();
                } catch (IOException ex) {
                    ex.printStackTrace();
                }
            }
            wait.dispose();
            if (mc3 instanceof Model) {
                new UIMessage(true, "Model fully loaded ! " + mc3, null);
                return mc3;
            } else {
                new UIMessage(true, "Error on loading ! " + filename + " - see log output.", null);
                return null;
            }
        }
    }

    public boolean isCompositeEnabled() {
        return compositeEnabled;
    }

    public void setCompositeEnabled(boolean b) {
        compositeEnabled = b;
    }

    public Composite getComposite() {
        return cps;
    }

    public void setComposite(Composite cps) {
        this.cps = cps;
    }

    public Paint getPaint() {
        return pnt;
    }

    public void setPaint(Paint pnt) {
        this.pnt = pnt;
    }

    public Color getColor() {
        return clr;
    }

    public void setColor(Color clr) {
        this.clr = clr;
    }

    public boolean isInnerResourceModeEnabled() {
        return _innerRsrcMode;
    }

    public void setInnerResourceModeEnabled(boolean b) {
        _innerRsrcMode = b;
    }

    public boolean isMultiThreadingEnabled() {
        return false;
    }

    public void setMultiThreadingEnabled(boolean b) {
    }

    public Object loadResource() {
        try {
            load(true, new SpritesCacheAdapter() {

                public void capacityExtended(int newCapacity) {
                    System.out.println("cache extended to " + newCapacity);
                }

                public void readCompleted() {
                    System.out.println("Model read data complete.");
                }

                public void writeCompleted() {
                    System.out.println("Model write data complete.");
                }
            });
            return null;
        } catch (InterruptedException ex) {
            ex.printStackTrace();
            Thread.currentThread().interrupt();
            return null;
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }

    public boolean isResourceLoaded() {
        return state == LOADED;
    }

    /**
    @see ThreadGroup#getParent()*/
    public ThreadGroup getThreadGroup() {
        return validateMonitor.getParent();
    }

    /***/
    public void setThreadGroup(ThreadGroup tg) {
        painting = false;
        validating = false;
        if (validateMonitor instanceof ThreadGroup) {
            validateMonitor.interrupt();
        }
        if (paintMonitor instanceof ThreadGroup) {
            paintMonitor.interrupt();
        }
        if (tg instanceof ThreadGroup) {
            animSynch = new CoalescedThreadsMonitor(tg, "TG-" + getClass().getName() + "-anim " + id);
            validateMonitor = new CoalescedThreadsMonitor(tg, "TG-" + getClass().getName() + "-pack " + id);
            paintMonitor = new CoalescedThreadsMonitor(tg, "TG-" + getClass().getName() + "-paint " + id);
        } else {
            animSynch = new CoalescedThreadsMonitor("TG-" + getClass().getName() + "-anim " + id);
            validateMonitor = new CoalescedThreadsMonitor("TG-" + getClass().getName() + "-pack " + id);
            paintMonitor = new CoalescedThreadsMonitor("TG-" + getClass().getName() + "-paint " + id);
        }
        validateMonitor.setMaxPriority(Thread.MAX_PRIORITY);
        validateMonitor.setCoalesceEnabled(false);
        paintMonitor.setMaxPriority(validateMonitor.getMaxPriority() - 1);
        paintMonitor.setCoalesceEnabled(true);
    }
}
