package net.library.jiga.sf3.game;
import java.awt.im.InputContext;
import java.io.File;
import java.io.Serializable;
import java.awt.*;
import java.awt.event.KeyEvent;
import java.util.*;
import javax.swing.JComponent;
import javax.swing.JTextField;
import net.library.jiga.sf3.game.physics.Fighter;
import net.library.jiga.sf3.system.input.KeyEventWrapper;

/**
 * SpritesChars classes.
 * frames are the anims start (0) and end sprites (1).
 * anims are named as follows specifically: forward is from the left to the right, as player one.
 * throw is from the left to the right, player one throws player two.
 * thrown is from the right to the left, player one is thrown by player two.
 * backwards and blocks from the right to the left, as player one.
 * reverse lists all the anims to be reversed, see above.
*/

public abstract class SpritesChar extends JComponent implements Comparable, Serializable{   
    /** character's id */
    protected String id = "";
    /** height between bottom of pics to bottom of the sprites in pixels */
    protected int floorDiff = 0;
    /** width of the pics */
    protected int spriteWidth = 0;
    /** height of the pics*/
    protected int spriteHeight = 0;
    /** average internal size of the sprites in the pics */
    protected Dimension spriteSize = new Dimension(0,0);
    protected int[][] frames = new int[][]{{0,0}};
    protected String[] reversedAnims = new String[]{""};
    
    protected String[] sfx = new String[]{""};
    /***/
    public String animsID_default = "stance";
    public String[] animsID = Fighter._playerActions;
    public String prefix = "";
    public String suffix = "";
    public static Map<String, InteractiveModel> characters;
    
    public final static int LOW = 0;
    public final static int MID = 1;
    public final static int HIGH = 2;
    public final static int[] RES = new int[]{LOW,MID,HIGH};
    public final static double LOW_FACTOR = 0.33;
    public final static double MID_FACTOR = 0.66;
    public final static double HIGH_FACTOR = 1.0;
    public final static String[] XPNG = new String[]{"image/x-png", ".x-png"};
    public final static String[] PNG = new String[]{"image/png", ".png"};
    public final static String[] GIF = new String[]{"image/gif", ".gif"};
    public final static String[] BMP = new String[]{"image/bmp", ".bmp"};
    public final static String[] JPEG = new String[]{"image/jpeg", ".jpg"};
    public final static HashMap<Integer, String> RES_NAMES_map = new HashMap<Integer, String>();
    public final static HashMap<Integer, Double> RES_FACTORS_map = new HashMap<Integer, Double>();
    public final static HashMap<Integer, String> RES_PATHS_map = new HashMap<Integer, String>();
    public final static HashMap<Integer, String> RES_EXT_map = new HashMap<Integer, String>();
    public final static HashMap<Integer, String> RES_MIMETYPES_map = new HashMap<Integer, String>();
    static{
        RES_NAMES_map.put(LOW, "low resolution");
        RES_NAMES_map.put(MID, "middle resolution");
        RES_NAMES_map.put(HIGH, "high resolution");
        RES_FACTORS_map.put(LOW, LOW_FACTOR);
        RES_FACTORS_map.put(MID, MID_FACTOR);
        RES_FACTORS_map.put(HIGH, HIGH_FACTOR);
        RES_PATHS_map.put(LOW, "low");
        RES_PATHS_map.put(MID, "mid");
        RES_PATHS_map.put(HIGH, "");
        RES_EXT_map.put(LOW, XPNG[1]);
        RES_EXT_map.put(MID, XPNG[1]);
        RES_EXT_map.put(HIGH, PNG[1]);
        RES_MIMETYPES_map.put(LOW, XPNG[0]);
        RES_MIMETYPES_map.put(MID, XPNG[0]);
        RES_MIMETYPES_map.put(HIGH, PNG[0]);
    }
    /** file resolution to sprites images
     * @see #RES_EXT_map
     * @see #RES_MIMETYPES_map"*/
    public int res;
    
    /***/
    public static String DEFAULT_EXT = PNG[0];
    
    /***/
    public final static Map<String, Integer> keyCodeMap = Fighter._keyCodeMap;
    
    /** gets the internal directory path to Sprites @ desired resolution
     * @param res LOW, MID or HIGH
     * @return path */
    public String getPath(boolean rsrcMode, int res) {
        String resPath = null;
        switch(res) {
            case LOW:
                resPath = RES_PATHS_map.get(LOW);
                break;
            case HIGH:
                resPath = RES_PATHS_map.get(HIGH);
                break;
            case MID:
                resPath = RES_PATHS_map.get(MID);
                break;
            default:
                break;
        }
        return id + ((resPath != null)?((rsrcMode)?"/":File.separator) + resPath:"");
    }
    
    public static Custom getCustomSpritesChar(int res) {
        return new Custom(res);
    }
    
    public static Custom getCustomSpritesChar(String name, int res, int floorDiff, Dimension spriteDim, Dimension spriteSize, int[][] framesNum, String[] reversedAnims, String[] sfxRsrcs) {
        return new Custom(name, res, floorDiff, spriteDim, spriteSize, framesNum, reversedAnims, sfxRsrcs);
    }
    
    public String toString() {
        return id;
    }
    
    public int compareTo(Object o) {
        if(o instanceof SpritesChar) {
            return id.compareTo(((SpritesChar)o).id);
        }
        return 1;
    }
    
    public boolean equals(Object o) {
        if(compareTo(o) == 0)
            return true;
        else
            return false;
    }
    
    public int anim(String name) {return 0;}
    
    public int sound(String name) {return 0;}
    
    public void setResolution(int res) {
        this.res = res;
    }
    
    public int getResolution() {
        return res;
    }
    
    public int getFloorDiff(int spritesRes) {
        return (int)((double)floorDiff * RES_FACTORS_map.get(spritesRes));
    }
    
    public int getWidth(int spritesRes) {
        return (int)((double)spriteWidth * RES_FACTORS_map.get(spritesRes));
    }
    
    public int getHeight(int spritesRes) {
        return (int)((double)spriteHeight * RES_FACTORS_map.get(spritesRes));
    }
    
    static {
        characters = new HashMap<String, InteractiveModel>();
        InteractiveModel model;
        String[] sfx;
        // Chunli
        sfx = new String[17];
        for(int i = 0; i < sfx.length; i++)
            sfx[i] = "sf3-sfx " + (i+23) + ".aif";
        Vector<KeyEventWrapper> seq;
        characters.put("ChunLi",
                new InteractiveModel("chunli",
                HIGH,
                30,
                new Dimension(576,230),
                new Dimension(85, 105),
                new int[][] {{19457,19461},{19472,19479}, {19488,19505},
                {19520,19537}, {19552,19583}, {19584,19590}, {19591,19599}, {19600,19605}, {19606,19615}, {19616,19641},
                {19642,19647}, {19648,19655}, {19656,19663}, {19664,19680}, {19681,19689}, {19696,19704}, {19712,19722},
                {19728,19736}, {19744,19788}, {19792,19793}, {19795,19818}, {19819,19827}, {19829,19839},
                {19840,19848}, {19856,19864}, {19872,19883}, {19888,19896}, {19904,19913}, {19920,19926}, {19936, 19957},
                {19959,19979}, {19984,19994}, {20000,20025}, {20032,20037}, {20048,20062}, {20064,20076}, {20080,20099}, {20112,20115}, {20128,20133},
                {20144,20160}, {20176,20178}, {20192,20195}, {20208,20277}, {20288,20292}, {20304,20314}, {20320,20332}, {20336,20344},
                {20352,20358}, {20368,20382}, {20384,20394}, {20400,20407}, {20416,20427}, {20432,20442}, {20448,20462},
                {20464,20472}, {20480,20497}, {20512,20518}, {20528,20532}, {20544,20552}, {20560,20566}, {20576,20580},
                {20592,20600}, {20608,20618}, {20624,20636}, {20640,20659}, {20672,20675}, {20688,20700}, {20704,20713},
                {20720,20728}, {20734,20766},
                {20768,20798}, {20800,20814}, {20816,20837}, {20848,20861}, {20864,20876}, {20880,20939}, {20944,20985},
                {20992,21099}},
                new String[]{"jump forwards", "forwards", "forwards LP", "forwards MP", "forwards HP",
                "forwards LK", "forwards MK", "forwards HK", "jump forwards LK"},
                sfx
                )
                );
        model = characters.get("ChunLi");
        model.setAttribute("player", "ONE");
        model.addKeyEventWrapperMapping(new KeyEventWrapper(-1, -1, 3000), "");
        model.addAnimMap("intro", 0);
        model.addAnimMap("stance", 1);
        model.addAnimMap("forwards", 2);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED), "forwards");
        model.addAnimMap("backwards", 3);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED), "backwards");
        model.addAnimMap("crouch", 5);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED), "crouch");
        model.addAnimMap("win", 6);
        model.addAnimMap("draw", 8);
        model.addAnimMap("high recover", 12);
        model.addAnimMap("H block", 14);
        model.addAnimMap("M block", 15);
        model.addAnimMap("L block", 16);
        model.addAnimMap("jump", 17);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED), "jump");
        model.addAnimMap("H hit", 24);
        model.addAnimMap("M hit", 25);
        model.addAnimMap("L hit", 26);
        model.addAnimMap("super hit", 28);
        model.addAnimMap("super hit", 31);
        model.addAnimMap("forwards LP", 41);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards LP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards LP");
        model.addAnimMap("forwards MP", 42);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards MP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards MP");
        model.addAnimMap("forwards HP", 43);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards HP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards HP");
        model.addAnimMap("LP", 44);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LP"), KeyEvent.KEY_PRESSED), "LP");
        model.addAnimMap("MP", 45);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MP"), KeyEvent.KEY_PRESSED), "MP");
        model.addAnimMap("HP", 46);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HP"), KeyEvent.KEY_PRESSED), "HP");
        model.addAnimMap("jump backwards", 47);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump backwards");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump backwards");
        model.addAnimMap("forwards LK", 48);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards LK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards LK");
        model.addAnimMap("forwards MK", 49);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards MK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards MK");
        model.addAnimMap("forwards HK", 50);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards HK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards HK");
        model.addAnimMap("LK", 51);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED), "LK");
        model.addAnimMap("MK", 52);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED), "MK");
        model.addAnimMap("HK", 53);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED), "HK");
        model.addAnimMap("jump forwards", 47);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump forwards");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump forwards");
        model.addAnimMap("M hit", 24);
        model.addAnimMap("H hit", 25);
        model.addAnimMap("L hit", 26);
        model.addAnimMap("back hit", 27);
        model.addAnimMap("L hit", 30);
        model.addAnimMap("super hit", 31);
        model.addAnimMap("thrown", 32);
        model.addAnimMap("LP", 44);
        model.addAnimMap("crouch LP", 54);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch LP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch LP");
        model.addAnimMap("crouch MP", 55);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch MP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch MP");
        model.addAnimMap("crouch HP", 56);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch HP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch HP");
        model.addAnimMap("crouch LK", 57);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch LK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch LK");
        model.addAnimMap("crouch MK", 58);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch MK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch MK");
        model.addAnimMap("crouch HK", 59);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch HK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch HK");
        model.addAnimMap("H block", 14);
        model.addAnimMap("M block", 15);
        model.addAnimMap("L block", 16);
        model.addAnimMap("jump LP", 60);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump LP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump LP");
        model.addAnimMap("jump MP", 61);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump MP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump MP");
        model.addAnimMap("jump HP", 62);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump HP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump HP");
        model.addAnimMap("jump LK", 63);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump LK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump LK");
        model.addAnimMap("jump MK", 64);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump MK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump MK");
        model.addAnimMap("jump HK", 65);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump HK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump HK");
        model.addAnimMap("throw", 69);
        model.addAnimMap("jump throw", 70);
        model.addAnimMap("jump forwards LK", 71);
        model.addAnimMap("super 2", 72);
        model.addAnimMap("super 1", 73);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED, 2000));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_RELEASED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "super 1");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED, 2000));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_RELEASED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "super 1");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED, 2000));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_RELEASED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "super 1");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED, 2000));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_RELEASED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "super 1");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED, 2000));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_RELEASED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "super 1");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED, 2000));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_RELEASED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "super 1");
        model.addAnimMap("super 3", 74);
        
        model.addSoundMap("draw", 8);
        model.addSoundMap("win", 15);
        model.addSoundMap("super 1", 0);
        model.addSoundMap("fireball 1", 1);
        model.addSoundMap("super 2", 4);
        model.addSoundMap("intro", 7);
        model.addSoundMap("feign", 11);
        model.addSoundMap("lose", 16);
        
        //demo
        characters.put("ChunLi-demo",
                new InteractiveModel("chunli",
                HIGH,
                30,
                new Dimension(576,230),
                new Dimension(85, 105),
                new int[][] {{19457,19461}},
                new String[]{"jump forwards", "forwards", "forwards LP", "forwards MP", "forwards HP",
                "forwards LK", "forwards MK", "forwards HK", "jump forwards LK"},
                sfx
                )
                );
        model = characters.get("ChunLi-demo");
        model.setAttribute("player", "ONE");
        model.addKeyEventWrapperMapping(new KeyEventWrapper(-1, -1, 3000), "");
        model.addAnimMap("intro", 0);
        model.addAnimMap("stance", 0);
        model.addAnimMap("forwards", 0);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED), "forwards");
        model.addAnimMap("backwards", 0);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED), "backwards");
        model.addAnimMap("crouch", 0);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED), "crouch");
        model.addAnimMap("win", 0);
        model.addAnimMap("draw", 0);
        model.addAnimMap("high recover", 0);
        model.addAnimMap("H block", 0);
        model.addAnimMap("M block", 0);
        model.addAnimMap("L block", 0);
        model.addAnimMap("jump", 0);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED), "jump");
        model.addAnimMap("H hit", 0);
        model.addAnimMap("M hit", 0);
        model.addAnimMap("L hit", 0);
        model.addAnimMap("super hit", 0);
        model.addAnimMap("super hit", 0);
        model.addAnimMap("forwards LP", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards LP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards LP");
        model.addAnimMap("forwards MP", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards MP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards MP");
        model.addAnimMap("forwards HP", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards HP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards HP");
        model.addAnimMap("LP", 0);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LP"), KeyEvent.KEY_PRESSED), "LP");
        model.addAnimMap("MP", 0);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MP"), KeyEvent.KEY_PRESSED), "MP");
        model.addAnimMap("HP", 0);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HP"), KeyEvent.KEY_PRESSED), "HP");
        model.addAnimMap("jump backwards", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump backwards");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump backwards");
        model.addAnimMap("forwards LK", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards LK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards LK");
        model.addAnimMap("forwards MK", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards MK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards MK");
        model.addAnimMap("forwards HK", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards HK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "forwards HK");
        model.addAnimMap("LK", 0);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED), "LK");
        model.addAnimMap("MK", 0);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED), "MK");
        model.addAnimMap("HK", 0);
        model.addKeyEventWrapperMapping(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED), "HK");
        model.addAnimMap("jump forwards", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump forwards");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump forwards");
        model.addAnimMap("M hit", 0);
        model.addAnimMap("H hit", 0);
        model.addAnimMap("L hit", 0);
        model.addAnimMap("back hit", 0);
        model.addAnimMap("L hit", 0);
        model.addAnimMap("super hit", 0);
        model.addAnimMap("thrown", 0);
        model.addAnimMap("LP", 0);
        model.addAnimMap("crouch LP", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch LP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch LP");
        model.addAnimMap("crouch MP", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch MP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch MP");
        model.addAnimMap("crouch HP", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch HP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch HP");
        model.addAnimMap("crouch LK", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch LK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch LK");
        model.addAnimMap("crouch MK", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch MK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch MK");
        model.addAnimMap("crouch HK", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch HK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-down"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "crouch HK");
        model.addAnimMap("H block", 0);
        model.addAnimMap("M block", 0);
        model.addAnimMap("L block", 0);
        model.addAnimMap("jump LP", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump LP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump LP");
        model.addAnimMap("jump MP", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump MP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump MP");
        model.addAnimMap("jump HP", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HP"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump HP");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HP"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump HP");
        model.addAnimMap("jump LK", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump LK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump LK");
        model.addAnimMap("jump MK", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump MK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump MK");
        model.addAnimMap("jump HK", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump HK");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-up"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "jump HK");
        model.addAnimMap("throw", 0);
        model.addAnimMap("jump throw", 0);
        model.addAnimMap("jump forwards LK", 0);
        model.addAnimMap("super 2", 0);
        model.addAnimMap("super 1", 0);
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED, 2000));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_RELEASED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "super 1");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED, 2000));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_RELEASED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-LK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "super 1");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED, 2000));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_RELEASED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "super 1");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED, 2000));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_RELEASED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-MK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "super 1");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED, 2000));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_RELEASED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "super 1");
        seq = new Vector<KeyEventWrapper>();
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_PRESSED, 2000));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-backwards"), KeyEvent.KEY_RELEASED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-HK"), KeyEvent.KEY_PRESSED));
        seq.add(new KeyEventWrapper(keyCodeMap.get(model.getAttribute("player") + "-forwards"), KeyEvent.KEY_PRESSED));
        model.addKeyEventWrapperMapping(new KeyEventWrapper(seq), "super 1");
        model.addAnimMap("super 3", 0);
        
        model.addSoundMap("draw", 8);
        model.addSoundMap("win", 15);
        model.addSoundMap("super 1", 0);
        model.addSoundMap("fireball 1", 1);
        model.addSoundMap("super 2", 4);
        model.addSoundMap("intro", 7);
        model.addSoundMap("feign", 11);
        model.addSoundMap("lose", 16);
        
        //Ken
        sfx = new String[11];
        for(int i = 0; i < sfx.length; i++)
            sfx[i] = "sf3-sfx " + (i+12) + ".aif";
        characters.put("Ken",
                new InteractiveModel("ken",
                HIGH,
                50,
                new Dimension(368, 238),
                new Dimension(80, 110),
                new int[][]{{14657,14669}, {14672,14682}, {14684,14694}, {14696,14707}, {14714,14717}, {14720,14764},
                {14768,14779},
                {14784,14796}, {14800,14814}, {14816,14824}, {14826,14839}, {14842,14864}, {14880,14908}, {14911,14932}, {14944,14998},
                {15008,15037}, {15040,15067}, {15073,15078}, {15104,15110}, {15112,15132}, {15138,15148}, {15152,15156}, {15164,15172},
                {15184,15205}, {15216,15222}, {15233,15240}, {15246,15254}, {15264,15288}, {15296,15318}, {15328, 15359}, {15368,15372}, {15376,15388},
                {15408,15417}, {15438,15438}, {15440,15449}, {15456,15469}, {15536,15552}, {15568,15584}, {15601,15623},
                {15626,15628},
                {15632,15646}, {15648,15658}, {15680,15688}, {15696,15706}},
                null,
                sfx
                ));
        model = characters.get("Ken");
        model.addAnimMap("stance", 0);
        model.addAnimMap("forwards MP", 13);
        
        /*Ryu
        sfx = new String[12]; sfx[0] = "sf3-sfx.aif";
        for(int i = 1; i < sfx.length; i++)
            sfx[i] = "sf3-sfx " + (i) + ".aif";
        characters.put("Ryu",
                new InteractiveModel("ryu",
                HIGH,
                27,
                new Dimension(290,178),
                new Dimension(80, 100),
                new int[][]{{2593,2605}, {2608,2618}, {2620,2630}, {2632,2643},{2650,2653}, {2656,2700},
                {2704,2715}, {2720,2732}, {2736,2750}, {2752,2760}, {2762,2775}, {2778,2800}, {2816,2844}, {2847,2868},
                {2880,2931}, {2933,2935}, {2944,2973}, {2976,3003}, {3016,3038}, {3040,3046}, {3048,3053}, {3065,3073},
                {3075,3081}, {3088,3092}, {3100,3108}, {3120,3141}, {3152,3167}, {3169,3176}, {3182,3190}, {31200,3224},
                {3232,3254}, {3264,3308}, {3312,3325}, {3344,3374}, {3376,3386}, {3392,3442}, {3456,3464}, {3472,3486},
                {3488,3495}, {3504,3514}, {3520,3530}},
                null,
                sfx
                ));
        model = characters.get("Ryu");
        model.prefix = "0";
        */
        /*Elena
        sfx = new String[1];
        characters.put("Elena",
                new InteractiveModel("elena",
                HIGH,
                106,
                new Dimension(350, 275),
                new Dimension(100, 100),
                new int[][]{{10208,10265}, {10266,10276}, {10277,10286}, {10287,10293}, {10294,10296}, {10297,10304}, {10305,10311},
                {10312,10332}, {10333,10340}, {10341,10344}, {10345,10348}, {10349,10365}, {10366,10381}, {10382,10391}, {10392,10401},
                {10402,10419}, {10420,10424}, {10425,10446}, {10447,10463}, {10464,10475}, {10476,10481}, {10482,10488}, {10489,10496},
                {10497,10504}, {10505,10517}, {10518,10532}, {10533,10547}, {10548,10552}, {10553,10565}, {10566,10584}, {10585,10596},
                {10597,10603}, {10604,10615}, {10616,10621}, {10622,10628}, {10629,11757}},
                null,
                sfx
                ));
        */
        /*Makoto
        sfx = new String[1];
        characters.put("Makoto",
                new InteractiveModel("makoto",
                HIGH,
                43,
                new Dimension(406, 215),
                new Dimension(100, 100),
                new int[][]{{21121,22986}},
                null,
                sfx
                ));
         * */
    }
}
