/*
 * Fighter.java
 *
 * Created on 20 juillet 2007, 04:31
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package net.library.jiga.sf3.game.physics;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Rectangle2D;
import java.util.Collections;
import java.util.Iterator;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Vector;
import net.library.jiga.sf3.game.InteractiveModel;

/**
 * TODO Model management
 * @author website
 * Class that acts like a Thread to compute the player's new position for each key or combinaison pressed (MOVE_UP,MOVE_DOWN,... or nothing)
 * The fighter is positioned with the bottom of its actual sprite display. --------
 * |		|
 * |		|
 * |	x	|  <- sprite
 * |		|
 * |		|
 * ----.---
 *
 * ^- O(x,y)
 */
public class Fighter extends Player {

    private final static Vector<String> _actions = new Vector<String>();

    static {
        _actions.add("stance");
        _actions.add("intro");
        _actions.add("forwards");
        _actions.add("forwards LP");
        _actions.add("forwards MP");
        _actions.add("forwards HP");
        _actions.add("forwards LK");
        _actions.add("forwards MK");
        _actions.add("forwards HK");
        _actions.add("backwards");
        _actions.add("backwards LP");
        _actions.add("backwards MP");
        _actions.add("backwards HP");
        _actions.add("backwards LK");
        _actions.add("backwards MK");
        _actions.add("backwards HK");
        _actions.add("jump");
        _actions.add("jump LP");
        _actions.add("jump MP");
        _actions.add("jump HP");
        _actions.add("jump LK");
        _actions.add("jump MK");
        _actions.add("jump HK");
        _actions.add("crouch");
        _actions.add("crouch LP");
        _actions.add("crouch MP");
        _actions.add("crouch HP");
        _actions.add("crouch LK");
        _actions.add("crouch MK");
        _actions.add("crouch HK");
        _actions.add("jump forwards");
        _actions.add("jump forwards LP");
        _actions.add("jump forwards MP");
        _actions.add("jump forwards HP");
        _actions.add("jump forwards LK");
        _actions.add("jump forwards MK");
        _actions.add("jump forwards HK");
        _actions.add("jump backwards");
        _actions.add("jump backwards LP");
        _actions.add("jump backwards MP");
        _actions.add("jump backwards HP");
        _actions.add("jump backwards LK");
        _actions.add("jump backwards MK");
        _actions.add("jump backwards HK");
        _actions.add("LP");
        _actions.add("MP");
        _actions.add("HP");
        _actions.add("LK");
        _actions.add("MK");
        _actions.add("HK");
        _actions.add("throw");
        _actions.add("jump throw");
        _actions.add("H block");
        _actions.add("M block");
        _actions.add("L block");
        _actions.add("super block");
        _actions.add("H hit");
        _actions.add("M hit");
        _actions.add("L hit");
        _actions.add("back hit");
        _actions.add("super hit");
        _actions.add("thrown");
        _actions.add("recover");
        _actions.add("high recover");
        _actions.add("win");
        _actions.add("lose");
        _actions.add("draw");
        _actions.add("super 1");
        _actions.add("super 2");
        _actions.add("super 3");
        _actions.add("turn");
        _actions.add("turn crouching");
        _actions.add("fall");
        _actions.add("feign");
        _actions.add("fireball 1");
        _actions.add("fireball 2");
        _actions.add("fireball 3");
        _actions.add("K.O.");
    }
    public final static String[] _playerActions;

    static {
        _playerActions = new String[_actions.size()];
        for (int i = 0; i < _playerActions.length; i++) {
            _playerActions[i] = _actions.get(i);
        }
    }
    int life, lifeInertia;
    int ballSpeed, weight, speed, gravity;
    /***/
    Fighter[] opponents;
    /***/
    Rectangle fieldBounds;
    /***/
    SortedMap<String, SortedMap<Integer, Rectangle2D>> collisionBounds = Collections.synchronizedSortedMap(new TreeMap<String, SortedMap<Integer, Rectangle2D>>());

    /**
     *
     * @param opponents this array contains all present opponents fighters
     * @param fieldBounds
     * @param playerSize
     * @param floor
     * @param playerForwards
     * @param nextPlayerAction
     * @param lastPlayerAction
     * @param attributes
     * @param life
     * @param lifeInertia
     * @param pos
     * @param ballSpeed
     * @param fighterSpeed
     * @param fighterWeight
     * @param gravity
     */
    public Fighter(InteractiveModel model, Fighter[] opponents, Rectangle fieldBounds,
            SortedMap<String, SortedMap<Integer, Rectangle2D>> collisionBounds,
            int life, int lifeInertia, Point pos,
            int ballSpeed, int fighterSpeed, int fighterWeight,
            int gravity) {
        super(model, pos);
        setState(model, opponents, fieldBounds, collisionBounds,
                life, lifeInertia, pos, ballSpeed,
                fighterSpeed, fighterWeight,
                gravity);
    }

    /***/
    public void setState(InteractiveModel model, Fighter[] opponents, Rectangle fieldBounds,
            SortedMap<String, SortedMap<Integer, Rectangle2D>> collisionBounds,
            int life, int lifeInertia, Point pos,
            int ballSpeed, int fighterSpeed, int fighterWeight,
            int gravity) {
        super.setState(model, pos);
        synchronized (collisionBounds) {
            for (Iterator<String> i = collisionBounds.keySet().iterator(); i.hasNext();) {
                String key = i.next();
                this.collisionBounds.put(key, collisionBounds.get(key));
            }
        }
        this.fieldBounds = fieldBounds;
        this.opponents = opponents;
        this.life = life;
        this.lifeInertia = lifeInertia;
        this.ballSpeed = ballSpeed;
        this.speed = fighterSpeed;
        this.weight = fighterWeight;
        this.gravity = gravity;
    }

    /***/
    private boolean hasOpponentsAtLeft() {
        for (Fighter f : opponents) {
            if (f.pos.x < pos.x) {
                return true;
            }
        }
        return false;
    }

    /***/
    private boolean hasOpponentsAtRight() {
        for (Fighter f : opponents) {
            if (f.pos.x > pos.x) {
                return true;
            }
        }
        return false;
    }

    /***/
    public void resetTicker() {
        last_tick = 0L;
    }

    //Computes the player position and inertia
    //the inertia: actually the acceleration is computed, but the speed
    public void run() {
        current_tick = System.currentTimeMillis();
        if (last_tick <= 0) {
            last_tick = current_tick;
        }
        System.out.println(toString() + " : (\\) next action " + model.getAnimName());
        int xDirection = (hasOpponentsAtRight() && !hasOpponentsAtLeft()) ? 1 : ((!hasOpponentsAtRight() && hasOpponentsAtLeft()) ? -1 : 1);
        if (hasOpponentsAtLeft() && hasOpponentsAtRight()) {
            Point op = null;
            for (Fighter f : opponents) {
                if (op != null) {
                    op = f.pos;
                }
                op = (Math.abs(pos.x - f.pos.x) < Math.abs(pos.x - op.x)) ? f.pos : op;
            }
            xDirection = (op.x < pos.x) ? -1 : 1;
        }
        System.out.println("(|) (x,y) calculation");
        int id = _actions.indexOf(model.getAnimName());
        if (id == _actions.indexOf("forwards")) {
            calcStraightRun(xDirection, speed);
        } else if (id == _actions.indexOf("backwards")) {
            calcStraightRun(-xDirection, speed);
        } else if (id == _actions.indexOf("jump")) {
            if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(0, t, speed);
            lock(model, _actions.get(id));
        } else if (id == _actions.indexOf("jump backwards")) {
            if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(-xDirection, t, speed);
            lock(model, _actions.get(id));
        } else if (id == _actions.indexOf("jump forwards")) {
            if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(xDirection, t, speed);
            lock(model, _actions.get(id));
        } else if (id == _actions.indexOf("stance")) {
            resetJump();
        } else if(id == _actions.indexOf("jump LP")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(0, t, speed);
            lock(model, "jump");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump MP")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(0, t, speed);
            lock(model, "jump");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump HP")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(0, t, speed);
            lock(model, "jump");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump LK")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(0, t, speed);
            lock(model, "jump");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump MK")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(0, t, speed);
            lock(model, "jump");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump HK")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(0, t, speed);
            lock(model, "jump");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump forwards LP")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(xDirection, t, speed);
            lock(model, "jump forwards");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump forwards MP")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(xDirection, t, speed);
            lock(model, "jump forwards");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump forwards HP")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(xDirection, t, speed);
            lock(model, "jump forwards");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump forwards LK")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(xDirection, t, speed);
            lock(model, "jump forwards");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump forwards MK")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(xDirection, t, speed);
            lock(model, "jump forwards");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump forwards HK")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(xDirection, t, speed);
            lock(model, "jump forwards");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump backwards LP")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(-xDirection, t, speed);
            lock(model, "jump backwards");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump backwards MP")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(-xDirection, t, speed);
            lock(model, "jump backwards");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump backwards HP")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(-xDirection, t, speed);
            lock(model, "jump backwards");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump backwards LK")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(-xDirection, t, speed);
            lock(model, "jump backwards");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump backwards MK")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(-xDirection, t, speed);
            lock(model, "jump backwards");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        } else if(id == _actions.indexOf("jump backwards HK")) {
             if (t == 0) {
                t = System.currentTimeMillis();
            }
            calcHighJump(-xDirection, t, speed);
            lock(model, "jump backwards");
            model.anim(_actions.get(id), model.getFrameRate(), true, false);
        }
            
        if (falling) {
            if (model.getAnimName().contains("backwards")) {
                calcFall((Math.PI - 1) / alphaJumpSlope * xDirection, t);
            } else if (model.getAnimName().contains("forwards")) {
                calcFall(-(Math.PI - 1) / alphaJumpSlope * xDirection, t);
            } else {
                calcFall(-(Math.PI) / alphaJumpSlope, t);
            }
        }
        // collisions
        Rectangle2D collision = calcCollision();
        if (collision instanceof Rectangle2D) {
            int x = 0;
            if (collision.getX() > getBounds2D().getX()) { // collision on the right
                x = (int) (-1 + collision.getX() - ((float) getBounds2D().getWidth() / 2f));
            } else if (collision.getX() == getBounds2D().getX()) { // total collision, don't move !
                x = pos.x;
            } else { // collision on the left
                x = (int) (1 + collision.getMaxX() + ((float) getBounds2D().getWidth() / 2f));
            }
            int y = 0;
            if (collision.getY() > getBounds2D().getY()) { // collision on bottom
                y = -1 + (int) collision.getY();
            } else if (collision.getY() == getBounds2D().getY()) { // total collision, don't move !
                y = pos.y;
            } else { // collision on top
                y = 1 + (int) (collision.getMaxY() + getBounds2D().getHeight());
            }
            pos.x = x;
            pos.y = y;
            // replace in allowed bounds
            pos.x = Math.min(fieldBounds.x + fieldBounds.width - 1, Math.max(fieldBounds.x + 1, pos.x));
            pos.y = Math.min(fieldBounds.y + fieldBounds.height - 1, Math.max(fieldBounds.y + 1, pos.y));
            unlock(model);
            model.anim(model.animsID_default, model.getFrameRate(), true, false);
            if (pos.y < fieldBounds.y + fieldBounds.height - 1) {
                if (model.getAnimName().contains("backwards")) {
                    calcFall((Math.PI - 1) / alphaJumpSlope * xDirection, t = current_tick);
                } else if (model.getAnimName().contains("forwards")) {
                    calcFall(-(Math.PI - 1) / alphaJumpSlope * xDirection, t = current_tick);
                } else {
                    calcFall(Math.PI / alphaJumpSlope, t = current_tick);
                }
            } else {
                resetJump();
            }
        }
        last_tick = System.currentTimeMillis();
        System.out.println("(/) " + toString() + "'s new POSITION (x,y)=" + pos.x + "," + pos.y + " on field : " + fieldBounds);
    }
    /***/
    boolean falling = false;
    /***/
    Point air_start = null;
    /***/
    long t = 0;

    /***/
    public void resetJump() {
        air_start = null;
        falling = false;
        t = 0L;
    }

    /***/
    private void calcFall(double xDirection, long start) {
        if (start == current_tick) {
            resetJump();
        }
        falling = true;
        calcHighJump(xDirection, start, (int) ((float) speed / 2f));
    }
    /***/
    private double alphaJumpSlope = Math.PI / 16.0;
    /* F = m.a ; a(t) = v'(t) = x''(t);
    x(t) = x0 + v0 * t * cos(alpha) ; y(t) = y0 + v0 * t * sin(alpha) - 1/2 G * t^2
    angle is PI/2 - (xDirection * alphaJumpSlope)
     */

    private void calcHighJump(double xDirection, long start, int speed) {
        if (air_start == null) {
            air_start = pos;
        }
        double alpha = Math.PI / 2.0 - (double) xDirection * alphaJumpSlope;
        float elapsed = (float) ((float) (current_tick - start) / 1000f);
        //System.out.println("/*************** " + "height=" + (max.y - pos.y) + " sin(pi/4)=" + Math.sin(alpha) + " y-move=" + (Math.sin(alpha) * speed + 0.9 * t));
        pos.x = air_start.x + (int) ((float) Math.cos(alpha) * (float) speed * elapsed);
        pos.y = air_start.y - (int) ((float) Math.sin(alpha) * (float) speed * elapsed - .5f * (float) gravity * Math.pow(elapsed, 2.0));
    }

    /***/
    public Rectangle2D calcCollision() {
        Rectangle2D collision;
        for (Fighter f : opponents) {
            collision = f.getBounds2D().createIntersection(getBounds2D());
            if (!collision.isEmpty()) {
                return collision;
            }
        }
        Rectangle2D top, right, bottom, left;
        int boundsThickness = fieldBounds.width;
        top = new Rectangle(fieldBounds.x - boundsThickness, fieldBounds.y - boundsThickness, fieldBounds.width + (int) (2f * (float) boundsThickness), boundsThickness);
        right = new Rectangle(fieldBounds.x + fieldBounds.width, fieldBounds.y, boundsThickness, fieldBounds.height);
        bottom = new Rectangle(fieldBounds.x - boundsThickness, fieldBounds.y + fieldBounds.height, fieldBounds.width + (int) (2f * (float) boundsThickness), boundsThickness);
        left = new Rectangle(fieldBounds.x - boundsThickness, fieldBounds.y, boundsThickness, fieldBounds.height);
        if ((!(collision = getBounds2D().createIntersection(top)).isEmpty()) ||
                (!(collision = getBounds2D().createIntersection(left)).isEmpty()) ||
                (!(collision = getBounds2D().createIntersection(bottom)).isEmpty()) ||
                (!(collision = getBounds2D().createIntersection(right)).isEmpty())) {
            return collision;
        } else {
            return null;
        }
    }

    /***/
    private void calcStraightRun(int xDirection, int speed) {
        pos.x += (int) ((float) xDirection / Math.abs((float) xDirection) * (float) speed * (float) ((float) (current_tick - last_tick) / 1000f));
        pos.x = Math.min(Math.max(fieldBounds.x, pos.x), (int) fieldBounds.getMaxX());
    }

    /** computes the inertia as defined in run() */
    private int calcInertie() {
        int s;
        if (lifeInertia < 0) {
            s = -1;
        } else {
            s = 1;
        }
        return s * (int) Math.pow(Math.abs(lifeInertia), 2);
    }

    public int getLifeInertia() {
        return lifeInertia;
    }
}
