/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package net.library.jiga.sf3.game.physics;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.KeyEvent;
import java.awt.geom.Rectangle2D;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import net.library.jiga.sf3.game.InteractiveModel;
import net.java.games.input.*;
/**
 *
 * @author tianarakoto
 */
public abstract class Player implements Runnable {    
    /***/
    public final static Map<String, Integer> _keyCodeMap = Collections.synchronizedMap(new HashMap<String, Integer>());

    static {
        _keyCodeMap.put("ONE-forwards", KeyEvent.VK_RIGHT);
        _keyCodeMap.put("ONE-down", KeyEvent.VK_DOWN);
        _keyCodeMap.put("ONE-backwards", KeyEvent.VK_LEFT);
        _keyCodeMap.put("ONE-up", KeyEvent.VK_UP);
        _keyCodeMap.put("TWO-forwards", KeyEvent.VK_A);
        _keyCodeMap.put("TWO-down", KeyEvent.VK_S);
        _keyCodeMap.put("TWO-backwards", KeyEvent.VK_D);
        _keyCodeMap.put("TWO-up", KeyEvent.VK_W);
        _keyCodeMap.put("ONE-start", KeyEvent.VK_1);
        _keyCodeMap.put("TWO-start", KeyEvent.VK_2);
        _keyCodeMap.put("ONE-LP", KeyEvent.VK_SPACE);//KeyEvent.VK_U);
        _keyCodeMap.put("ONE-MP", KeyEvent.VK_ALT);//KeyEvent.VK_I);
        _keyCodeMap.put("ONE-HP", KeyEvent.VK_CONTROL);//KeyEvent.VK_O);
        _keyCodeMap.put("ONE-LK", KeyEvent.VK_BACK_SPACE);//KeyEvent.VK_J);
        _keyCodeMap.put("ONE-MK", KeyEvent.VK_ENTER);//KeyEvent.VK_K);
        _keyCodeMap.put("ONE-HK", KeyEvent.VK_SHIFT);//KeyEvent.VK_L);
        _keyCodeMap.put("TWO-LP", KeyEvent.VK_R);
        _keyCodeMap.put("TWO-MP", KeyEvent.VK_T);
        _keyCodeMap.put("TWO-HP", KeyEvent.VK_Z);
        _keyCodeMap.put("TWO-LK", KeyEvent.VK_F);
        _keyCodeMap.put("TWO-MK", KeyEvent.VK_G);
        _keyCodeMap.put("TWO-HK", KeyEvent.VK_H);
    }
    
    /***/
    public final static Map<String, String> _gamepadCodeMap = Collections.synchronizedMap(new HashMap<String, String>());
    
    static{
        
        _gamepadCodeMap.put("ONE-forwards", Component.Identifier.Axis.X.getName());
        _gamepadCodeMap.put("ONE-down", Component.Identifier.Axis.Y.getName());
        _gamepadCodeMap.put("ONE-backwards", Component.Identifier.Axis.X.getName());
        _gamepadCodeMap.put("ONE-up", Component.Identifier.Axis.Y.getName());
        _gamepadCodeMap.put("TWO-forwards", Component.Identifier.Axis.X.getName());
        _gamepadCodeMap.put("TWO-down", Component.Identifier.Axis.Y.getName());
        _gamepadCodeMap.put("TWO-backwards", Component.Identifier.Axis.X.getName());
        _gamepadCodeMap.put("TWO-up", Component.Identifier.Axis.Y.getName());
        _gamepadCodeMap.put("ONE-start", Component.Identifier.Button._9.getName());
        _gamepadCodeMap.put("TWO-start", Component.Identifier.Button._9.getName());
        _gamepadCodeMap.put("ONE-LP", Component.Identifier.Button._0.getName());
        _gamepadCodeMap.put("ONE-MP", Component.Identifier.Button._1.getName());
        _gamepadCodeMap.put("ONE-HP", Component.Identifier.Button._2.getName());
        _gamepadCodeMap.put("ONE-LK", Component.Identifier.Button._3.getName());
        _gamepadCodeMap.put("ONE-MK", Component.Identifier.Button._4.getName());
        _gamepadCodeMap.put("ONE-HK", Component.Identifier.Button._5.getName());
        _gamepadCodeMap.put("TWO-LP", Component.Identifier.Button._0.getName());
        _gamepadCodeMap.put("TWO-MP", Component.Identifier.Button._1.getName());
        _gamepadCodeMap.put("TWO-HP", Component.Identifier.Button._2.getName());
        _gamepadCodeMap.put("TWO-LK", Component.Identifier.Button._3.getName());
        _gamepadCodeMap.put("TWO-MK", Component.Identifier.Button._4.getName());
        _gamepadCodeMap.put("TWO-HK", Component.Identifier.Button._5.getName());
    }
    /***/
    long last_tick = 0L;
    /***/
    long current_tick = 0L;
    InteractiveModel model;
    Point pos;
    /***/
    Player(InteractiveModel model, Point pos) {
        setState(model, pos);
    }
    
    void setState(InteractiveModel model, Point pos) { 
        this.model = model;
        this.pos = pos;
    }
    
    public void run() {
        current_tick = System.currentTimeMillis();
        // do some actions
        System.out.println(getClass().getName() + this + " is being run.");
        last_tick = current_tick;
    }
    
    void resetTicker() {
        last_tick = 0L;
    }
    
    public Point getPosPlayer() {
        return pos;
    }

    public Rectangle2D getBounds2D() {
        return new Rectangle(pos.x - (int) ((float) model.getInsetSize().width / 2.0f), pos.y - model.getInsetSize().height, model.getInsetSize().width, model.getInsetSize().height);
    }

    /***/
    public String toString() {
        return getClass().getName() + "-" + model.getAttribute("name");
    }

    void lock(InteractiveModel model, String animName) {
        model.anim(animName, model.getFrameRate(), true, true);
    }

    void unlock(InteractiveModel model) {
        model.anim(model.getAnimName_Lock(), model.getFrameRate(), true, false);
    }
}
